#include <GL/glew.h>
#include <GL/gl.h>
#include <GLFW/glfw3.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "shader.h"
#include "log.h"
#include "shader.h"
#include "matrix-math.h"

#define STATUS_INTERVAL 0.5
#define PI 3.14159f

GLFWwindow* window;

GLuint program;
GLuint originProgram;
GLuint floorProgram;

GLuint vertexArrayObject;

GLuint originVAO;

GLuint indexBuffer;
GLuint lineIndexBuffer;
GLuint floorIndexBuffer;

GLuint initialWindowWidth = 800;
GLuint initialWindowHeight = 600;

GLfloat aspect;
mat4 realProjectionMatrix;

GLfloat currentTime = 0.0f;

int frameCount = 0;
struct timespec last_time, current_time;

typedef struct ColorRGB {
	GLfloat r;
	GLfloat g;
	GLfloat b;
} ColorRGB;

// Color Conversion Functions from https://gist.github.com/ciembor/1494530
GLfloat hueToRgb(GLfloat p, GLfloat q, GLfloat t) {
	if (t < 0) t += 1;
	if (t > 1) t -= 1;
	if (t < 1./6) return p + (q - p) * 6 * t;
	if (t < 1./2) return q;
	if (t < 2./3) return p + (q - p) * (2./3 - t) * 6;
	
  return p;
}

// Color Conversion Functions from https://gist.github.com/ciembor/1494530
ColorRGB hslToRgb(GLfloat h, GLfloat s, GLfloat l) {
	ColorRGB result = {0, 0, 0};

	if(0 == s) {
		result.r = result.g = result.b = l; // achromatic
	} else {
		float q = l < 0.5 ? l * (1 + s) : l + s - l * s;
		float p = 2 * l - q;
		result.r = hueToRgb(p, q, h + 1./3);
		result.g = hueToRgb(p, q, h);
		result.b = hueToRgb(p, q, h - 1./3);
	}

	return result;
}

void initialiseStatusDisplay() {
	clock_gettime(CLOCK_MONOTONIC, &last_time);
}

void updateStatusDisplay() {
	frameCount++;
	clock_gettime(CLOCK_MONOTONIC, &current_time);
	double elapsed = (current_time.tv_sec - last_time.tv_sec) +
					 (current_time.tv_nsec - last_time.tv_nsec) / 1e9;
	if (elapsed >= STATUS_INTERVAL) {
		double fps = frameCount / elapsed;
		frameCount = 0;
		last_time = current_time;
		printf("\rFPS: %5.2f - currentTime: %5.2f", fps, currentTime);
		fflush(stdout);
	}
}

void recalculateProjectionMatrix() {
	mat4BuildPerspective(realProjectionMatrix, 90 * M_PI / 180, aspect, 0.1, 50);
	DEBUG("Recalculating Projection Matrix");
}

void init(void) {

	INFO("Building main program...");
	ProgramLinkResult linkResult = buildShaderProgram("src/shaders/vertex.glsl", "src/shaders/fragment.glsl");
	if (!linkResult.success) {
		FATAL("Failed to link Program");
		exit(1);
	}

	program = linkResult.program;


	INFO("Building origin program...");
	linkResult = buildShaderProgram("src/shaders/origin-vertex.glsl", "src/shaders/origin-fragment.glsl");
	if (!linkResult.success) {
		FATAL("Failed to link Program");
		exit(1);
	}

	originProgram = linkResult.program;

	INFO("Building floor program...");
	linkResult = buildShaderProgram("src/shaders/floor-vertex.glsl", "src/shaders/floor-fragment.glsl");
	if (!linkResult.success) {
		FATAL("Failed to link Program");
		exit(1);
	}

	floorProgram = linkResult.program;

	INFO("Shader Program Done.");

	// create triangle buffer
	/**
	 * -0.35              0.35
	 * |     -0.2    0.2  |
	 * |    |        |    |
	 * 
	 * 0----1        4----5    ---  0.6
	 * |    |        |    |    
	 * |    |        |    |
	 * |    2--------3    |    ---  0.1
	 * |                  |
	 * |    9--------8    |    --- -0.1
	 * |    |        |    |
	 * |    |        |    |
	 * 11---10       7----6    --- -0.6
	 *
	 * 12-----------------13   --- -0.7
	 * |                  |
	 * 15-----------------14    --- -0.9
	 *   
	 */    

	GLfloat vertices[] = {
		// X  // Y
		 1.0f,  1.0f,  1.0f,
		 1.0f,  1.0f, -1.0f,
		 1.0f, -1.0f,  1.0f,
		 1.0f, -1.0f, -1.0f,
		-1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f, -1.0f,
		-1.0f, -1.0f,  1.0f,
		-1.0f, -1.0f, -1.0f,
	};

	GLuint restart = 128;
	glEnable(GL_PRIMITIVE_RESTART);
	glPrimitiveRestartIndex(restart);

	GLuint indices[] = {
		1, 0, 2,
		1, 2, 3,

		4, 0, 1,
		5, 4, 1,

		5, 1, 7,
		1, 3, 7,

		5, 7, 6,
		4, 5, 6,

		0, 4, 6,
		2, 0, 6,

		3, 2, 6,
		7, 3, 6
	};

	GLuint lineIndices[] = {
		0, 1,
		0, 2,
		0, 4,

		1, 3,
		1, 5,

		2, 3,
		2, 6,

		3, 7,

		4, 5,
		4, 6,

		5, 7,

		6, 7
	};

	GLuint floorIndices[] = {
		2, 3, 6,
		3, 7, 6
	};

	DEBUG("Creating vertext buffer");
	GLuint vertexBuffer;
	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	DEBUG("Creating vertex array object");
	// create vertex array object
	glGenVertexArrays(1, &vertexArrayObject);
	glBindVertexArray(vertexArrayObject);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);

	// vertex position data
	glVertexAttribPointer(
		0, // shader location
		3, // number of values to read
		GL_FLOAT, // type of value
		GL_FALSE, // if values are normalised
		3 * sizeof(GLfloat), // stride - distance between vertices
		0 // start offset
	);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(
		1, // shader location
		3, // number of values to read
		GL_FLOAT, // type of value
		GL_FALSE, // if values are normalised
		3 * sizeof(GLfloat), // stride - distance between vertices
		0 // start offset
	);
	glEnableVertexAttribArray(1);

	DEBUG("Creating index buffers");


	glGenBuffers(1, &indexBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);



	glGenBuffers(1, &lineIndexBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, lineIndexBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(lineIndices), lineIndices, GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);



	glGenBuffers(1, &floorIndexBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, floorIndexBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(floorIndices), floorIndices, GL_STATIC_DRAW);


	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glBindVertexArray(0);


	glClearColor(0.3f, 0.3f, 0.4f, 1.0f);
	glViewport(0, 0, initialWindowWidth, initialWindowHeight);

	// glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);

	glEnable(GL_DEPTH_TEST);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	DEBUG("Creating origin vertex buffer");
	GLfloat originVertices[] = {
	//  X     Y     Z     R     G     B
		0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
		1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f,
		0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f
	};
	GLuint originVBO;
	glGenBuffers(1, &originVBO);
	glBindBuffer(GL_ARRAY_BUFFER, originVBO);
	glBufferData(GL_ARRAY_BUFFER, sizeof(originVertices), originVertices, GL_STATIC_DRAW);

	glGenVertexArrays(1, &originVAO);
	glBindVertexArray(originVAO);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(GLfloat), 0);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(GLfloat), (void*)(3 * sizeof(GLfloat)));
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	initialiseStatusDisplay();

	recalculateProjectionMatrix();

	INFO("--- Initialisation done ---");
}

void drawCube(mat4 modelMatrix, mat4 viewMatrix) {
	mat4 modelViewMatrix;
	mat4Multiply(modelViewMatrix, viewMatrix, modelMatrix);

	// send modelView and projection matrix to shader
	GLuint modelViewLocation = glGetUniformLocation(program, "uModelView");
	glUniformMatrix4fv(modelViewLocation, 1, GL_FALSE, modelViewMatrix);

	// draw cube
	glBindVertexArray(vertexArrayObject);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer);
	glDrawElements(GL_TRIANGLES, 48, GL_UNSIGNED_INT, 0);
}

GLfloat clamp(GLfloat d, GLfloat min, GLfloat max) {
	const GLfloat t = d < min ? min : d;
	return t > max ? max : t;
}

GLfloat mapTo01(GLfloat d, GLfloat min, GLfloat max) {
	return (clamp(d, min, max) - min) * (1 / (max - min));
}

void draw(void) {
	// 0.0: virtual view
	// 0.1 transition to real view
	// 0.15 done
	// 0.3 start view transformation
	// 0.35 done
	// 0.4 start projection
	// 0.45 done
	// 0.55 - repeat
	updateStatusDisplay();

	mat4 identity;
	mat4Identity(identity);

	mat4 projectionMatrix;
	// mat4Interpolate(projectionMatrix, identity, projectionMatrix, mapTo01(currentHue, 0.0f, 0.1f));

	// counter for animation
	currentTime += 0.0001f;
	int state = glfwGetKey(window, GLFW_KEY_SPACE);
	if (state == GLFW_PRESS)
	{
		currentTime += 0.001f;
	}
	if (currentTime > 0.55f) currentTime = 0.0f;

	// clear colour and depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	// build view matrix
	mat4 viewMatrix;

	// mat4Identity(viewMatrix);
	vec3 cameraPos = {cos(currentTime * 6 * M_PI) * 3.0f, 3.0f + sin(currentTime * 12 * M_PI) * 0.5, sin(currentTime * 6 * M_PI) * 3.0f};
	vec3 cameraLookAt = {0.0f, 0.0f, 0.0f};
	vec3 cameraUp = {0.0f, 1.0f, 0.0f};
	mat4BuildLookAt(viewMatrix, cameraPos, cameraLookAt, cameraUp);


	vec3 virtualCameraPosition = {1.5f, 1.0f, 1.0f};
	vec3 virtualCameraLookAt = {0.0f, 0.0f, -1.0f};

	// calculate virtual lookAt and projection
	mat4 virtualViewMatrix;
	mat4 virtualProjectionMatrix;
	mat4BuildLookAt(virtualViewMatrix, virtualCameraPosition, virtualCameraLookAt, cameraUp);
	mat4BuildPerspective(virtualProjectionMatrix, 60 * M_PI / 180, aspect, 1, 4);

	mat4 inverseVirtualView;
	mat4Inverse(inverseVirtualView, virtualViewMatrix);
	mat4Interpolate(inverseVirtualView, inverseVirtualView, identity, mapTo01(currentTime, 0.3f, 0.35f));

	mat4 inverseVirtualProjection;
	mat4Inverse(inverseVirtualProjection, virtualProjectionMatrix);

	mat4 viewFrustumModelView;
	mat4Multiply(viewFrustumModelView, inverseVirtualView, inverseVirtualProjection);

	mat3 mat3CameraRotationMatrix;
	mat3From4(mat3CameraRotationMatrix, virtualViewMatrix);
	mat3Inverse(mat3CameraRotationMatrix, mat3CameraRotationMatrix);
	mat4 cameraRotationMatrix;
	mat4From3(cameraRotationMatrix, mat3CameraRotationMatrix);

	mat4Interpolate(projectionMatrix, virtualProjectionMatrix, realProjectionMatrix, mapTo01(currentTime, 0.1f, 0.15f));
	mat4Interpolate(viewMatrix, virtualViewMatrix, viewMatrix, mapTo01(currentTime, 0.1f, 0.15f));

	mat4Interpolate(virtualViewMatrix, identity, virtualViewMatrix, mapTo01(currentTime, 0.3f, 0.35f));
	mat4Interpolate(virtualProjectionMatrix, identity, virtualProjectionMatrix, mapTo01(currentTime, 0.4f, 0.45f) * mapTo01(currentTime, 0.4f, 0.45f));


	glUseProgram(program);

	GLuint brightnessLocation = glGetUniformLocation(program, "uBrightness");

	GLuint projectionLocation = glGetUniformLocation(program, "uProjection");
	glUniformMatrix4fv(projectionLocation, 1, GL_FALSE, projectionMatrix);

	// cubes.
	mat4 modelMatrix;
	vec3 scale;
	vec3 position;

	// cube 1
	mat4Identity(modelMatrix);

	vec3Set(scale, 0.5f, 0.5f, 0.5f);
	mat4Scale(modelMatrix, modelMatrix, scale);

	mat4RotateY(modelMatrix, modelMatrix, 0.5f);

	vec3Set(position, -0.3f, 0.5f, -0.4f);
	mat4Translate(modelMatrix, modelMatrix, position);

	mat4Multiply(modelMatrix, virtualViewMatrix, modelMatrix);
	mat4Multiply(modelMatrix, virtualProjectionMatrix, modelMatrix);

	glUniform1f(brightnessLocation, 1.0f);
	drawCube(modelMatrix, viewMatrix);

	// cube 2
	mat4Identity(modelMatrix);

	vec3Set(scale, 0.3f, 0.3f, 0.3f);
	mat4Scale(modelMatrix, modelMatrix, scale);

	mat4RotateY(modelMatrix, modelMatrix, 0.5f);

	vec3Set(position, -0.9f, 0.3f, -3.8f);
	mat4Translate(modelMatrix, modelMatrix, position);

	mat4Multiply(modelMatrix, virtualViewMatrix, modelMatrix);
	mat4Multiply(modelMatrix, virtualProjectionMatrix, modelMatrix);

	glUniform1f(brightnessLocation, 1.0f);
	drawCube(modelMatrix, viewMatrix);


	// draw virtual camera
	mat4Identity(modelMatrix);

	vec3Set(scale, 0.2f, 0.15f, 0.05f);
	mat4Scale(modelMatrix, modelMatrix, scale);

	// apply rotations to camera

	mat4Multiply(modelMatrix, cameraRotationMatrix, modelMatrix);

	mat4Translate(modelMatrix, modelMatrix, virtualCameraPosition);
	mat4Multiply(modelMatrix, virtualViewMatrix, modelMatrix);

	glUniform1f(brightnessLocation, 0.0f);
	
	if (currentTime < 0.37f) {
		drawCube(modelMatrix, viewMatrix);
	}



	glUseProgram(originProgram);

	projectionLocation = glGetUniformLocation(originProgram, "uProjection");
	glUniformMatrix4fv(projectionLocation, 1, GL_FALSE, projectionMatrix);

	GLuint modelViewLocation = glGetUniformLocation(originProgram, "uModelView");
	glUniformMatrix4fv(modelViewLocation, 1, GL_FALSE, viewMatrix);

	glBindVertexArray(originVAO);
	glLineWidth(5.0f);
	glDrawArrays(GL_LINES, 0, 6);

	if (currentTime < 0.65) {
		mat4Multiply(viewFrustumModelView, virtualProjectionMatrix, viewFrustumModelView);
		mat4Multiply(viewFrustumModelView, viewMatrix, viewFrustumModelView);

		glUniformMatrix4fv(modelViewLocation, 1, GL_FALSE, viewFrustumModelView);

		glBindVertexArray(vertexArrayObject);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, lineIndexBuffer);
		glDrawElements(GL_LINES, 24, GL_UNSIGNED_INT, 0);
	}

	// draw floor
	glUseProgram(floorProgram);
	glBindVertexArray(vertexArrayObject);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, floorIndexBuffer);

	glUniformMatrix4fv(glGetUniformLocation(floorProgram, "uView"), 1, GL_FALSE, viewMatrix);
	glUniformMatrix4fv(glGetUniformLocation(floorProgram, "uProjection"), 1, GL_FALSE, projectionMatrix);

	mat4 floorModelMatrix;
	vec3 floorScale;
	mat4Identity(floorModelMatrix);
	vec3Set(floorScale, 10.0f, 0.01f, 10.0f);
	mat4Scale(floorModelMatrix, floorModelMatrix, floorScale);

	glUniformMatrix4fv(glGetUniformLocation(floorProgram, "uModel"), 1, GL_FALSE, floorModelMatrix);

	glDepthMask(GL_FALSE);
	glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);
	glDepthMask(GL_TRUE);


}

void framebuffer_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);
	aspect = (float)width / height;

	recalculateProjectionMatrix();
}

int main(int argc, char const *argv[])
{

	for (int i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-v") == 0) {
			logLevel = LOG_LEVEL_INFO;
		}
		if (strcmp(argv[i], "-vv") == 0) {
			logLevel = LOG_LEVEL_DEBUG;
		}
	}

	INFO("Creating GLFW Window");

	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	window = glfwCreateWindow(initialWindowWidth, initialWindowHeight, "CG1", NULL, NULL);

	aspect = (float)initialWindowWidth / initialWindowHeight;
	if (!window) {
		FATAL("Failed to open window");
		glfwTerminate();
		return 1;
	}
	glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
	glfwMakeContextCurrent(window);

	glewInit();

	init();

	while (!glfwWindowShouldClose(window)) {
		draw();

		glfwSwapBuffers(window);
		glfwPollEvents();
	}

	glfwTerminate();

	return 0;
}